package billing

import (
	"context"
	"crypto/tls"
	"crypto/x509"
	"errors"
	"fmt"
	"io/ioutil"
	"strings"
	"time"

	"github.com/streadway/amqp"
)

type amqpClient struct {
	urls         []string
	exchange     string
	key          string
	deliveryMode uint8
	priority     uint8

	tlsConf  *tls.Config
	urlIndex int

	channelC chan *amqp.Channel
	doneC    chan struct{}
}

type amqpClientConfig struct {
	urls         string
	exchange     string
	key          string
	persistent   bool
	priority     uint
	caCertFile   string
	certFile     string
	keyFile      string
	depth        uint
	failIfNoPeer bool
	verify       bool
}

func newAMQPClient(conf *amqpClientConfig) (*amqpClient, error) {
	isSSL := conf.caCertFile != "" || conf.certFile != "" || conf.keyFile != ""

	if isSSL && conf.certFile != "" && conf.keyFile == "" {
		return nil, errors.New("A certificate file location was given, but no key file location")
	}
	if isSSL && conf.certFile == "" && conf.keyFile != "" {
		return nil, errors.New("A key file location was given, but no certificate file location")
	}

	urls := strings.Split(conf.urls, ",")
	for _, url := range urls {
		uri, err := amqp.ParseURI(url)
		if err != nil {
			return nil, fmt.Errorf("Creating AMQP client: %w", err)
		}

		if isSSL && uri.Scheme == "amqp" {
			log.Warnf(`SSL was enabled, but will not be used for amqp:// URL: "%s"`, url)
		}
	}

	if conf.exchange == "" {
		log.Notice("Billing AMQP exchange was not specified, default to using the default exchange")
	}
	if conf.priority > 9 {
		return nil, errors.New("Billing AMQP message priority cannot be higher than 9")
	}

	var deliveryMode uint8
	if conf.persistent {
		deliveryMode = amqp.Persistent
	} else {
		deliveryMode = amqp.Transient
	}

	tlsConf := new(tls.Config)
	tlsConf.RootCAs = x509.NewCertPool()

	if conf.caCertFile != "" {
		ca, err := ioutil.ReadFile(conf.caCertFile)
		if err != nil {
			return nil, fmt.Errorf("Reading CA cert file: %w", err)
		}
		tlsConf.RootCAs.AppendCertsFromPEM(ca)
	}

	if conf.certFile != "" {
		cert, err := tls.LoadX509KeyPair(conf.certFile, conf.keyFile)
		if err != nil {
			return nil, fmt.Errorf("Reading cert files: %w", err)
		}
		tlsConf.Certificates = append(tlsConf.Certificates, cert)
	}

	return &amqpClient{
		urls:         urls,
		exchange:     conf.exchange,
		key:          conf.key,
		deliveryMode: deliveryMode,
		priority:     uint8(conf.priority),
		tlsConf:      tlsConf,
		urlIndex:     0,
		channelC:     make(chan *amqp.Channel),
		doneC:        make(chan struct{}),
	}, nil
}

func (c *amqpClient) Start(ctx context.Context) {
	channelC := c.channelC
	doneC := ctx.Done()

connect:
	for {
		conn := c.connect(ctx)
		if conn == nil {
			// that means the context has been cancelled
			break connect
		}

		errC := make(chan *amqp.Error)
		conn.NotifyClose(errC)

	wait:
		for {
			select {
			case <-channelC:
				channel, err := conn.Channel()
				if err != nil {
					log.Errorf("Getting a channel: %s", err)
					channelC <- nil
				} else {
					channelC <- channel
				}

				continue wait
			case connErr := <-errC:
				url := c.urls[c.urlIndex]
				log.Errorf(`Connection to the broker "%s" has been lost, moving on to the next one: %s`, url, connErr)

				continue connect
			case <-doneC:
				if err := conn.Close(); err != nil {
					log.Warnf("Error when closing the connection to RabbitMQ: %s", err)
				}

				break connect
			}
		}
	}

	close(c.doneC)
}

func (c *amqpClient) connect(ctx context.Context) *amqp.Connection {
	doneC := ctx.Done()

	startIndex := c.urlIndex
	currentIndex := startIndex
	nbURLs := len(c.urls)
	backoffDuration := time.Millisecond
	var conn *amqp.Connection
	var err error
	for {
		select {
		case <-doneC:
			return nil
		case <-time.After(backoffDuration):
		}

		url := c.urls[currentIndex]

		conn, err = amqp.DialTLS(url, c.tlsConf)
		if err == nil {
			break
		}

		log.Errorf("Dialing AMQP queue: %s", err)

		currentIndex = (currentIndex + 1) % nbURLs

		if currentIndex == startIndex {
			backoffDuration *= 2

			if backoffDuration > time.Minute {
				backoffDuration = time.Minute
			}
		}
	}

	c.urlIndex = currentIndex

	return conn
}

func (c *amqpClient) Close() error {
	<-c.doneC

	return nil
}

func (c *amqpClient) Push(ctx context.Context, messageID string, message []byte) error {
	channelC := c.channelC
	doneC := ctx.Done()

	for {
		// get a channel
		var channel *amqp.Channel

		select {
		case channelC <- nil:
		case <-doneC:
			return fmt.Errorf("Cancelled: %w", ctx.Err())
		}
		for {
			select {
			case channel = <-channelC:
			case <-doneC:
				return fmt.Errorf("Cancelled: %w", ctx.Err())
			}

			if channel == nil {
				time.Sleep(100 * time.Millisecond)

				continue
			}

			break
		}
		defer channel.Close()

		// try to set the channel in confirmation mode
		var confirmC chan amqp.Confirmation
		if err := channel.Confirm(false); err != nil {
			log.Warn("Could not enable confirmation mode on channel: %s", err)
		} else {
			confirmC = make(chan amqp.Confirmation)
			channel.NotifyPublish(confirmC)
		}

		// push the message
		err := channel.Publish(
			c.exchange,
			c.key,
			true,  // mandatory
			false, // immediate
			amqp.Publishing{
				CorrelationId: messageID,
				Body:          message,
				DeliveryMode:  c.deliveryMode,
				Priority:      c.priority,
			})
		if err != nil {
			log.Errorf("Publishing to queue: %s", err)
			continue
		}

		if confirmC == nil {
			return nil
		}

		confirm := <-confirmC
		if !confirm.Ack {
			// no ack received, retry, pushing a duplicate is not a problem
			log.Errorf("Message was not acknowledged")

			continue
		}

		return nil
	}
}
