package oiofs

import (
	"context"
	"flag"
	"fmt"
	"strings"
	"time"

	"github.com/open-io/oio-exporter/common"
	_log "github.com/open-io/oio-exporter/common/log"
	"github.com/open-io/oio-exporter/daemon"
)

var log = _log.New("oiofs")

var confTimeout = 3 * time.Second
var confInterval = time.Duration(0)

func InitFlags(fs *flag.FlagSet) {
	fs.DurationVar(&confTimeout, "oiofs.timeout", confTimeout, "[oiofs] http timeout")
	fs.DurationVar(&confInterval, "oiofs.interval", confInterval, "[oiofs] interval between polls")
}

var Factory = &oiofsDaemonFactory{}

type oiofsDaemonFactory struct{}

func (df *oiofsDaemonFactory) Daemons(inventory *common.Inventory) map[string]interface{} {
	params := make(map[string]interface{})
	namespace := inventory.Namespace

	services := inventory.Services["oiofs"]
	for _, service := range services {
		params[service.ID] = &oiofsDaemonParams{
			Namespace:     namespace,
			ServiceID:     service.ID,
			ServiceIP:     service.IP,
			ServicePort:   service.Port,
			ServiceVolume: service.Volume,
		}
	}

	return params
}

func (df *oiofsDaemonFactory) New(paramsI interface{}) daemon.DaemonI {
	params := paramsI.(*oiofsDaemonParams)

	return &oiofsDaemon{
		namespace:     params.Namespace,
		serviceID:     params.ServiceID,
		serviceIP:     params.ServiceIP,
		servicePort:   params.ServicePort,
		serviceVolume: params.ServiceVolume,
	}
}

type oiofsDaemonParams struct {
	Namespace     string
	ServiceID     string
	ServiceIP     string
	ServicePort   int
	ServiceVolume string
}

type oiofsDaemon struct {
	daemon.BaseDaemon

	namespace     string
	serviceID     string
	serviceIP     string
	servicePort   int
	serviceVolume string
	serviceAddr   string
}

func (d *oiofsDaemon) Interval() time.Duration {
	return confInterval
}

func (d *oiofsDaemon) Init(ctx context.Context) error {
	d.serviceAddr = fmt.Sprintf("%s:%d", d.serviceIP, d.servicePort)

	return nil
}

func (d *oiofsDaemon) Collect(ctx context.Context, m *common.MetricData) error {
	target := d.serviceAddr

	url := fmt.Sprintf("http://%s/stats", target)
	requestCtx, cancel := context.WithTimeout(ctx, confTimeout)
	defer cancel()

	var data map[string]float64
	err := common.HTTPGetJson(requestCtx, url, &data)
	if err != nil {
		return err
	}

	for k, v := range data {
		k = strings.ToLower(k)
		k = strings.Replace(k, "init_ctx", "initCtx", 1)
		parts := strings.Split(k, "_")

		metric := ""
		switch parts[len(parts)-1] {
		case "us":
			metric = "operation_time"
		case "count":
			metric = "operation_count"
		case "byte":
			metric = "bytes"
		case "microseconds":
			metric = "chunk_age"
		case "hit", "miss":
			metric = "cache_read"
		case "succeeded", "failed":
			// TODO Do something with this
			continue
		default:
			// TODO This shouldn't happen, add debug log
			continue
		}
		m.Set(&common.Metric{
			Name: "oiofs_" + metric,
			Labels: map[string]string{
				"namespace":    d.namespace,
				"service_type": "oiofs",
				"service":      d.serviceID,
				"service_addr": target,
				"mountpoint":   d.serviceVolume,
				"operation":    strings.Trim(parts[1], " "),
				"dimension":    parts[2],
				"module":       parts[0],
			},
			Value: v,
			TTL:   confInterval * 2,
		})
	}

	return nil
}
